import * as d3 from 'd3';
import ELK from 'elkjs';


import { Module } from './layout';

export class NetlistRender {
    /**
     * 
     * @param {YosysRawNet} rawNet 
     * @param {number} renderHeight
     * @param {number} renderWidth
     */
    constructor() {
        /**
         * @type {ElkGraph}
         */
        this.elkGraph = {
            id: 'root',
            layoutOptions: {
                'elk.algorithm': 'layered',
                'layered.nodePlacement.strategy': 'SIDE_BASED'
            },
            children: [],
            edges: []
        };

        /**
         * @type {Map<string, Module>}
         */
        this.nameToModule = new Map();

    }

    /**
     * @description 加载 yosys 格式的 json
     * @param {YosysRawNet} rawNet 
     */
    load(rawNet) {
        /**
         * @type {YosysRawNet}
         */
        this.rawNet = rawNet;
        
        // 转换为 elkjs 格式的 graph
        for (const [moduleName, rawModule] of Object.entries(rawNet.modules)) {
            const top = parseInt(rawModule.attributes.top);
            // 一开始只渲染 top 模块
            if (top) {
                const module = new Module(moduleName, rawModule);
                const portNodes = module.makeNetsElkNodes();
                const cellNodes = module.makeCellsElkNodes();
                const [constantNodes, connectionEdges] = module.makeConnectionElkNodes();

                this.elkGraph.children.push(...portNodes);
                this.elkGraph.children.push(...cellNodes);
                this.elkGraph.children.push(...constantNodes);
                this.elkGraph.edges.push(...connectionEdges);

                this.nameToModule.set(moduleName, module);
            }
        }
    }

    /**
     * @description 根据信息创建布局对象
     * @returns {Promise<ElkNode>}
     */
    async createLayout() {
        const elk = new ELK();
        const graph = this.elkGraph;
        const layoutGraph = await elk.layout(graph);
        console.log(layoutGraph);

        return layoutGraph;
    }

    /**
     * 
     * @param {ElkNode} computedLayout 
     * @param {string} container 类似于 "#chart"
     * @returns {d3.Selection}
     */
    async render(computedLayout, container) {
        const virtualHeight = computedLayout.height;
        const virtualWidth = computedLayout.width;

        // 根据 height 进行放缩（可以通过设置进行调整）
        const ratio = this.renderHeight / virtualHeight;

        // 遍历计算布局进行创建
        const svg = d3.select(container).append('svg')
            .attr('width', virtualWidth)
            .attr('height', virtualHeight);
        
        await this.renderEntity(svg, computedLayout, ratio);
        await this.renderLine(svg, computedLayout, ratio);

        return svg;
    }

    /**
     * @description 绘制实体
     * @param {d3.Selection} svg 
     * @param {ElkNode} computedLayout 
     * @param {number} ratio
     */
    async renderEntity(svg, computedLayout, ratio) {
        // node 可能是如下的几类
        // - module 的 port
        // - 器件（基础器件 & 例化模块）
        // - 器件的 port

        // 生成用于绘制的 d3 数据结构
        const squares = [];
        for (const node of computedLayout.children) {            
            squares.push({
                x: node.x,
                y: node.y,
                width: node.width,
                height: node.height,
                fill: '#2D323B',
                text: node.renderName,
                rx: 3,
                ry: 3
            });

            // 如果存在 port，绘制 port
            for (const cellPort of node.ports || []) {
                squares.push({
                    x: cellPort.x + node.x,
                    y: cellPort.y + node.y,
                    width: cellPort.width,
                    height: cellPort.height,
                    fill: '#CB81DA',
                    text: '',
                    rx: 0,
                    ry: 0
                });    
            }
        }

        svg.selectAll('rect')
            .data(squares)
            .enter()
            .append('rect')
            .attr('x', data => data.x)
            .attr('y', data => data.y)
            .attr('width', data => data.width)
            .attr('height', data => data.height)
            .attr('fill', d => d.fill)
            .attr('stroke', '#CB81DA')
            .attr('stroke-width', 2)
            .attr('rx', d => d.rx)
            .attr('ry', d => d.ry);
    

        svg.selectAll('text')
            .data(squares)
            .enter()
            .append('text')
            .attr('x', data => data.x + data.width / 2) // 文本的 x 坐标（居中）
            .attr('y', data => data.y + data.height / 2) // 文本的 y 坐标（居中）
            .attr('dominant-baseline', 'middle') // 文本垂直居中
            .attr('text-anchor', 'middle') // 文本水平居中
            .attr('fill', 'white')              // 文本颜色
            .attr('font-size', '12px')
            .text(data => data.text); // 设置文本内容
    }

    /**
     * @description 绘制连线
     * @param {d3.Selection} svg 
     * @param {ElkNode} computedLayout 
     * @param {number} ratio
     */
    async renderLine(svg, computedLayout, ratio) {
        const lines = [];
        for (const edge of computedLayout.edges) {
            for (const section of edge.sections || []) {
                const points = [];
                points.push(section.startPoint);
                for (const point of section.bendPoints || []) {
                    points.push(point);
                }
                points.push(section.endPoint);

                for (let i = 0; i < points.length - 1; ++ i) {
                    lines.push({
                        x1: points[i].x,
                        y1: points[i].y,
                        x2: points[i + 1].x,
                        y2: points[i + 1].y,
                        strokeWidth: 2,
                        color: 'var(--foreground)'
                    });
                }
            }
        }

        svg.selectAll('line')
            .data(lines)
            .enter()
            .append('line')
            .attr('x1', data => data.x1)
            .attr('y1', data => data.y1)
            .attr('x2', data => data.x2)
            .attr('y2', data => data.y2)
            .attr('stroke-width', data => data.strokeWidth)
            .attr('stroke', data => data.color);
    }
}